from subprocess import run
from time import sleep

from ..abstractionhelper import is_installed, is_wayland
from .typer import Key, Typer


class XDoToolTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return not is_wayland() and is_installed("xdotool")

    @staticmethod
    def name() -> str:
        return "xdotool"

    def get_active_window(self) -> str:
        return run(args=["xdotool", "getactivewindow"], capture_output=True, encoding="utf-8").stdout[:-1]

    def type_characters(self, characters: str, start_delay: float, key_delay: int, active_window: str) -> None:
        sleep(start_delay)
        run(
            [
                "xdotool",
                "windowactivate",
                "--sync",
                active_window,
                "type",
                "--clearmodifiers",
                "--delay",
                str(key_delay),
                "--",
                characters,
            ]
        )
        # workaround for https://github.com/jordansissel/xdotool/issues/43
        run(["xdotool", "keyup", "Shift_L", "Shift_R", "Alt_L", "Alt_R"])

    def press_key(self, key: Key) -> None:
        match key:
            case Key.ENTER:
                key_name = "Return"
            case Key.TAB:
                key_name = "Tab"
            case _:
                raise Exception("Unknown key")

        run(["xdotool", "key", key_name])



from subprocess import run
from time import sleep

from ..abstractionhelper import is_installed, is_wayland
from .typer import Key, Typer


class WTypeTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("wtype")

    @staticmethod
    def name() -> str:
        return "wtype"

    def get_active_window(self) -> str:
        return "not possible with wtype"

    def type_characters(self, characters: str, start_delay: float, key_delay: int, active_window: str) -> None:
        sleep(start_delay)
        args = ["wtype"]

        if key_delay > 0:
            args = args + ["-d", str(key_delay)]

        args.append("--")
        args.append(characters)
        run(args)

    def press_key(self, key: Key) -> None:
        match key:
            case Key.ENTER:
                key_name = "return"
            case Key.TAB:
                key_name = "tab"
            case _:
                raise Exception("Unknown key")

        run(["wtype", "-k", key_name])



class Target:
    __value: str

    def __init__(self, input: str):
        prepared_input = input.strip()
        if prepared_input.lower() in (
            "brand",
            "cardholder",
            "code",
            "expiry",
            "menu",
            "notes",
            "number",
            "password",
            "totp",
            "username",
        ) or prepared_input.lower().startswith("uri"):
            self.__value = prepared_input.lower()
        else:
            self.__value = prepared_input

    def is_uri(self) -> bool:
        return self.__value.startswith("uri")

    def uri_index(self) -> int:
        if not self.is_uri():
            raise IndexError()

        if len(self.__value) == 3:
            return 0
        else:
            return int(self.__value[4:]) - 1

    @property
    def raw(self):
        return self.__value

    def __eq__(self, other):
        return self.__value == other.__value


class TypeTarget(Target):
    pass


class Targets:
    USERNAME = Target("username")
    PASSWORD = Target("password")
    TOTP = Target("totp")
    NUMBER = Target("number")
    CARDHOLDER = Target("cardholder")
    BRAND = Target("brand")
    EXPIRY = Target("expiry")
    CODE = Target("code")
    NOTES = Target("notes")
    MENU = Target("menu")


class TypeTargets:
    DELAY = TypeTarget("delay")
    ENTER = TypeTarget("enter")
    TAB = TypeTarget("tab")



from subprocess import run
from time import sleep

from ..abstractionhelper import is_installed
from .typer import Key, Typer


class DotoolTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_installed("dotool")

    @staticmethod
    def name() -> str:
        return "dotool"

    def get_active_window(self) -> str:
        return "not possible with dotool"

    def type_characters(self, characters: str, start_delay: float, key_delay: int, active_window: str) -> None:
        sleep(start_delay)
        run(["dotool"], input=f"typedelay {key_delay}\ntype {characters}", encoding="utf-8")

    def press_key(self, key: Key) -> None:
        match key:
            case Key.ENTER:
                key_name = "enter"
            case Key.TAB:
                key_name = "tab"
            case _:
                raise Exception("Unknown key")

        run(["dotool"], input=f"key {key_name}", encoding="utf-8")



from abc import ABC, abstractmethod
from dataclasses import dataclass, field

from rofi_rbw.models.field import Field
from rofi_rbw.models.targets import Target, TypeTarget


@dataclass(frozen=True)
class DetailedEntry(ABC):
    name: str
    folder: str | None = ""
    fields: list[Field] = field(default_factory=list)

    @abstractmethod
    def __getitem__(self, target: Target) -> str | None:
        pass

    @property
    @abstractmethod
    def default_target(self) -> list[Target]:
        pass

    @property
    @abstractmethod
    def default_autotype_target(self) -> list[TypeTarget]:
        pass

    @property
    def autotype_sequence(self) -> list[TypeTarget] | None:
        sequence = next((field for field in self.fields if field.key == "_autotype"), None)
        if sequence is None:
            return None

        return [TypeTarget(target_string) for target_string in sequence.value.strip().split(":")]



from subprocess import run

from ..abstractionhelper import is_installed
from ..models.action import Action
from ..models.detailed_entry import DetailedEntry
from ..models.display_field_token import DisplayFieldToken
from ..models.entry import Entry
from ..models.keybinding import Keybinding
from ..models.targets import Target
from .selector import Selector


class Bemenu(Selector):
    @staticmethod
    def supported() -> bool:
        return is_installed("bemenu")

    @staticmethod
    def name() -> str:
        return "bemenu"

    def show_selection(
        self,
        entries: list[Entry],
        prompt: str,
        show_help_message: bool,
        display_fields: list[DisplayFieldToken],
        keybindings: list[Keybinding],
        additional_args: list[str],
    ) -> tuple[None, Action | None, Entry | None]:
        parameters = ["bemenu", "-p", prompt, *additional_args]

        bemenu = run(
            parameters,
            input="\n".join(self._format_entries(entries, display_fields)),
            capture_output=True,
            encoding="utf-8",
        )
        if bemenu.returncode == 0:
            return None, None, self._find_entry(entries, bemenu.stdout, display_fields)
        else:
            return None, Action.CANCEL, None

    def select_target(
        self,
        entry: DetailedEntry,
        show_help_message: bool,
        keybindings: list[Keybinding],
        additional_args: list[str],
    ) -> tuple[list[Target] | None, Action | None]:
        parameters = ["bemenu", "-p", "Choose target", *additional_args]

        bemenu = run(
            parameters,
            input="\n".join(self._format_targets_from_entry(entry)),
            capture_output=True,
            encoding="utf-8",
        )

        if bemenu.returncode == 1:
            return None, Action.CANCEL

        return self._extract_targets(bemenu.stdout), None



import re
from abc import ABC, abstractmethod

from ..models.action import Action
from ..models.card import Card
from ..models.credentials import Credentials
from ..models.detailed_entry import DetailedEntry
from ..models.display_field_token import DisplayFieldToken
from ..models.entry import Entry
from ..models.keybinding import Keybinding
from ..models.note import Note
from ..models.targets import Target, Targets


class Selector(ABC):
    @staticmethod
    def best_option(name: str | None = None) -> "Selector":
        from .bemenu import Bemenu
        from .fuzzel import Fuzzel
        from .rofi import Rofi
        from .wofi import Wofi

        available_selectors = [Rofi, Wofi, Fuzzel, Bemenu]

        if name is not None:
            try:
                return next(selector for selector in available_selectors if selector.name() == name)()
            except StopIteration:
                raise NoSelectorFoundException()
        else:
            try:
                return next(selector for selector in available_selectors if selector.supported())()
            except StopIteration:
                raise NoSelectorFoundException()

    @staticmethod
    @abstractmethod
    def supported() -> bool:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @abstractmethod
    def show_selection(
        self,
        entries: list[Entry],
        prompt: str,
        show_help_message: bool,
        display_fields: list[DisplayFieldToken],
        keybindings: list[Keybinding],
        additional_args: list[str],
    ) -> tuple[list[Target] | None, Action | None, Entry | None]:
        pass

    @abstractmethod
    def select_target(
        self,
        entry: DetailedEntry,
        show_help_message: bool,
        keybindings: list[Keybinding],
        additional_args: list[str],
    ) -> tuple[list[Target] | None, Action | None]:
        pass

    def _format_entries(self, entries: list[Entry], display_fields: list[DisplayFieldToken]) -> list[str]:
        number_tokens = len(display_fields)
        formatted_entries = [[self._format_field(entry, token) for token in display_fields] for entry in entries]
        max_lengths = [max(len(field) for field in pivoted_fields) for pivoted_fields in zip(*formatted_entries)]

        return [
            "  ".join([entry[field_index].ljust(max_lengths[field_index]) for field_index in range(number_tokens)])
            for entry in formatted_entries
        ]

    def _find_entry(
        self, entries: list[Entry], formatted_string: str, display_fields: list[DisplayFieldToken]
    ) -> Entry:
        pattern = f"^{' {2,}'.join(f'(?P<{token.name.lower()}>.*?)' for token in display_fields)}$"
        match = re.compile(pattern).search(formatted_string)

        return next(
            entry
            for entry in entries
            if all(
                self._format_field(entry, token) == match.group(token.name.lower()).strip() for token in display_fields
            )
        )

    def _format_targets_from_entry(self, entry: DetailedEntry) -> list[str]:
        match entry:
            case Credentials():
                return self._format_targets_from_credential(entry)
            case Card():
                return self._format_targets_from_card(entry)
            case Note():
                return self._format_targets_from_note(entry)
            case _:
                return []

    def _format_targets_from_credential(self, credentials: Credentials) -> list[str]:
        targets = []
        if credentials.username:
            targets.append(f"Username: {credentials.username}")
        if credentials.password:
            targets.append(f"Password: {credentials.password[0]}{'*' * (len(credentials.password) - 1)}")
        if credentials.has_totp:
            targets.append(f"TOTP: {credentials.totp}")
        if credentials.notes:
            targets.append(f"Notes: {credentials.notes}")
        if len(credentials.uris) == 1:
            targets.append(f"URI: {credentials.uris[0]}")
        else:
            for key, value in enumerate(credentials.uris):
                targets.append(f"URI {key + 1}: {value}")
        for field in credentials.fields:
            targets.append(f"{self._format_further_item_name(field.key)}: {field.masked_string()}")

        return targets

    def _format_targets_from_card(self, card: Card) -> list[str]:
        targets = []
        if card.number:
            targets.append(f"Number: {card.number}")
        if card.cardholder_name:
            targets.append(f"Cardholder: {card.cardholder_name}")
        if card.brand:
            targets.append(f"Brand: {card.brand}")
        if card.exp_month and card.exp_year:
            targets.append(f"Expiry: {card.exp_year:0>4}-{card.exp_month:0>2}")
        if card.code:
            targets.append(f"Code: {card.code[0]}{'*' * (len(card.code) - 1)}")
        if card.notes:
            targets.append(f"Notes: {card.notes}")
        for field in card.fields:
            targets.append(f"{self._format_further_item_name(field.key)}: {field.masked_string()}")

        return targets

    def _format_targets_from_note(self, note: Note) -> list[str]:
        targets = []
        if note.notes:
            value = note.notes.replace("\n", "|")
            targets.append(f"Notes: {value}")
        for field in note.fields:
            targets.append(f"{self._format_further_item_name(field.key)}: {field.masked_string()}")

        return targets

    def _format_further_item_name(self, key: str) -> str:
        if key.lower() in ["username", "password", "totp"] or re.match(r"^URI \d+$", key):
            return f"{key} (field)"
        return key

    def _format_action_and_targets(self, keybinding: Keybinding) -> str:
        if keybinding.targets and Targets.MENU in keybinding.targets:
            return "Menu"
        elif keybinding.action == Action.SYNC:
            return "Sync logins"
        elif keybinding.targets:
            return f"{keybinding.action.value.title()} {', '.join([target.raw for target in keybinding.targets])}"
        else:
            return keybinding.action.value.title()

    def _format_field(self, entry: Entry, token: DisplayFieldToken) -> str:
        match token:
            case DisplayFieldToken.NAME_ONLY:
                return entry.name
            case DisplayFieldToken.NAME_WITH_FOLDER:
                return f"{entry.folder}/{entry.name}" if entry.folder else entry.name
            case DisplayFieldToken.FOLDER:
                return entry.folder
            case DisplayFieldToken.USER:
                return entry.username
            case DisplayFieldToken.FIRST_URI:
                return entry.uris[0] if entry.uris else ""

    def _extract_targets(self, output: str) -> list[Target]:
        return [Target(line.split(":")[0]) for line in output.strip().split("\n")]


class NoSelectorFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to show the selection. Please check the required dependencies."



from .clipboarder import Clipboarder, NoClipboarderFoundException


class NoopClipboarder(Clipboarder):
    @staticmethod
    def supported() -> bool:
        return True

    @staticmethod
    def name() -> str:
        return "noop"

    def copy_to_clipboard(self, characters: str) -> None:
        raise NoClipboarderFoundException()

    def clear_clipboard_after(self, clear: int) -> None:
        raise NoClipboarderFoundException()



import enum
from abc import ABC, abstractmethod


class Typer(ABC):
    @staticmethod
    def best_option(name: str | None = None) -> "Typer":
        from .dotool import DotoolTyper
        from .noop import NoopTyper
        from .wtype import WTypeTyper
        from .xdotool import XDoToolTyper
        from .ydotool import YDotoolTyper

        available_typers = [XDoToolTyper, WTypeTyper, YDotoolTyper, DotoolTyper, NoopTyper]

        if name is not None:
            return next(typer for typer in available_typers if typer.name() == name)()
        else:
            return next(typer for typer in available_typers if typer.supported())()

    @staticmethod
    @abstractmethod
    def supported() -> bool:
        pass

    @staticmethod
    @abstractmethod
    def name() -> str:
        pass

    @abstractmethod
    def get_active_window(self) -> str:
        pass

    @abstractmethod
    def type_characters(self, characters: str, start_delay: float, key_delay: int, active_window: str) -> None:
        pass

    @abstractmethod
    def press_key(self, key: "Key") -> None:
        pass


class NoTyperFoundException(Exception):
    def __str__(self) -> str:
        return "Could not find a valid way to type characters. Please check the required dependencies."


class Key(enum.Enum):
    ENTER = "enter"
    TAB = "tab"



import os
from pathlib import Path

if os.environ.get("XDG_CACHE_HOME"):
    cache_file = Path(os.environ.get("XDG_CACHE_HOME")) / "rofi-rbw.runcache"
else:
    cache_file = Path.home() / ".cache" / "rofi-rbw.runcache"

if os.environ.get("XDG_CONFIG_HOME"):
    config_home = Path(os.environ.get("XDG_CONFIG_HOME"))
else:
    config_home = Path.home() / ".config"

if os.environ.get("XDG_CONFIG_DIRS"):
    config_global = [Path(dir) for dir in os.environ.get("XDG_CONFIG_DIRS").split(":")]
else:
    config_global = [Path("/etc/xdg")]

config_file_locations = [str(directory / "rofi-rbw.rc") for directory in [config_home] + config_global]



from subprocess import run
from time import sleep

from ..abstractionhelper import is_installed, is_wayland
from .typer import Key, Typer


class YDotoolTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("ydotool")

    @staticmethod
    def name() -> str:
        return "ydotool"

    def get_active_window(self) -> str:
        return "not possible with ydotool"

    def type_characters(self, characters: str, start_delay: float, key_delay: int, active_window: str) -> None:
        sleep(start_delay)
        run(["ydotool", "type", "--key-delay", str(key_delay), characters])

    def press_key(self, key: Key) -> None:
        match key:
            case Key.ENTER:
                key_name = "28"
            case Key.TAB:
                key_name = "15"
            case _:
                raise Exception("Unknown key")

        run(["ydotool", "key", f"{key_name}:1", f"{key_name}:0"])



from dataclasses import dataclass

from .detailed_entry import DetailedEntry
from .targets import Target, Targets, TypeTarget, TypeTargets


@dataclass(frozen=True)
class Card(DetailedEntry):
    cardholder_name: str | None = None
    number: str | None = None
    brand: str | None = None
    exp_month: str | None = None
    exp_year: str | None = None
    code: str | None = None
    notes: str | None = None

    @property
    def default_target(self) -> list[Target]:
        return [Targets.NUMBER, Targets.EXPIRY, Targets.CODE]

    @property
    def default_autotype_target(self) -> list[TypeTarget]:
        return [Targets.NUMBER, TypeTargets.TAB, Targets.EXPIRY, TypeTargets.TAB, Targets.CODE]

    def __getitem__(self, target: Target) -> str | None:
        match target:
            case Targets.NUMBER:
                return self.number
            case Targets.CARDHOLDER:
                return self.cardholder_name
            case Targets.BRAND:
                return self.brand
            case Targets.EXPIRY:
                return f"{self.exp_year:0>4}-{self.exp_month:0>2}"
            case Targets.CODE:
                return self.code
            case _:
                return next(
                    (field.value for field in self.fields if field.key == target.raw.removesuffix(" (field)")), None
                )



from importlib import metadata
from importlib.metadata import PackageNotFoundError

try:
    __version__ = metadata.version(__package__)
except PackageNotFoundError:
    __version__ = "dirty"

del metadata



from dataclasses import dataclass, field
from subprocess import run

from .detailed_entry import DetailedEntry
from .field import Field
from .targets import Target, Targets, TypeTarget, TypeTargets


@dataclass(frozen=True)
class Credentials(DetailedEntry):
    username: str | None = ""
    password: str | None = ""
    has_totp: bool = False
    notes: str | None = ""
    uris: list[str] = field(default_factory=list)
    fields: list[Field] = field(default_factory=list)

    def __getitem__(self, target: Target) -> str | None:
        match target:
            case Targets.USERNAME:
                return self.username
            case Targets.PASSWORD:
                return self.password
            case Targets.TOTP:
                return self.totp
            case Targets.NOTES:
                return self.notes
            case _ if target.is_uri():
                return self.uris[target.uri_index()]
            case _:
                return next(
                    (field.value for field in self.fields if field.key == target.raw.removesuffix(" (field)")), None
                )

    @property
    def default_target(self) -> list[Target]:
        return [Targets.USERNAME, Targets.PASSWORD]

    @property
    def default_autotype_target(self) -> list[TypeTarget]:
        return [Targets.USERNAME, TypeTargets.TAB, Targets.PASSWORD]

    @property
    def totp(self):
        if not self.has_totp:
            return ""

        command = ["rbw", "code", self.name]
        if self.username:
            command.extend([self.username])
        if self.folder:
            command.extend(["--folder", self.folder])
        return run(command, capture_output=True, encoding="utf-8").stdout.strip()



from dataclasses import dataclass, field

from .detailed_entry import DetailedEntry
from .field import Field
from .targets import Target, Targets, TypeTarget


@dataclass(frozen=True)
class Note(DetailedEntry):
    notes: str | None = ""
    fields: list[Field] = field(default_factory=list)

    def __getitem__(self, target: Target) -> str | None:
        if target == Targets.NOTES:
            return self.notes
        else:
            return next(
                (field.value for field in self.fields if field.key == target.raw.removesuffix(" (field)")), None
            )

    @property
    def default_target(self) -> list[Target]:
        return [Targets.NOTES]

    @property
    def default_autotype_target(self) -> list[TypeTarget]:
        return [Targets.NOTES]



import argparse
import shlex
import sys

import configargparse

from . import __version__
from .models.action import Action
from .models.display_field_token import DisplayFieldToken
from .models.keybinding import Keybinding
from .models.targets import Target, TypeTarget
from .paths import config_file_locations


def parse_arguments(argv: list[str] | None = None) -> argparse.Namespace:
    parser = configargparse.ArgumentParser(
        description="Insert or copy passwords and usernames from Bitwarden using rofi or rofi-likes.",
        default_config_files=config_file_locations,
    )
    parser.add_argument("--version", action="version", version="rofi-rbw " + __version__)
    parser.add_argument(
        "--action",
        "-a",
        dest="action",
        action="store",
        choices=[action.value for action in Action],
        default=Action.TYPE.value,
        help="What to do with the selected entry",
    )
    parser.add_argument("--target", "-t", dest="targets", action="append", help="Which part of the entry do you want?")
    parser.add_argument(
        "--prompt", "-r", dest="prompt", action="store", default="Choose entry", help="Set rofi-rbw's prompt"
    )
    parser.add_argument(
        "--selector-args",
        dest="selector_args",
        action="store",
        default="",
        help="A string of arguments to give to the selector",
    )
    parser.add_argument(
        "--selector",
        dest="selector",
        action="store",
        type=str,
        choices=["rofi", "wofi", "fuzzel", "bemenu"],
        default=None,
        help="Choose the selector frontend",
    )
    parser.add_argument(
        "--clipboarder",
        dest="clipboarder",
        action="store",
        type=str,
        choices=["xsel", "xclip", "wl-copy"],
        default=None,
        help="Choose the application to access the clipboard with",
    )
    parser.add_argument(
        "--typer",
        dest="typer",
        action="store",
        type=str,
        choices=["xdotool", "wtype", "ydotool", "dotool"],
        default=None,
        help="Choose the application to type with",
    )
    parser.add_argument(
        "--clear-after",
        dest="clear",
        action="store",
        type=int,
        default=0,
        help="Limit the duration in seconds passwords stay in your clipboard. When not set or <= 0, passwords stay indefinitely",
    )
    parser.add_argument(
        "--no-help", dest="show_help", action="store_false", help="Don't show a help message about the shortcuts"
    )
    parser.add_argument(
        "--no-folder",
        dest="show_folders",
        action="store_false",
        help=argparse.SUPPRESS,
    )
    parser.set_defaults(show_folders=True)
    parser.add_argument(
        "--display-fields",
        dest="display_fields",
        action="store",
        nargs="+",
        choices=[field.value for field in DisplayFieldToken],
        default=None,
        metavar="DISPLAY_FIELDS",
        help='Ordered list of fields to show in the selection list, given as separate arguments Defaults to "name_with_folder user"',
    )
    parser.add_argument("--no-cache", dest="use_cache", action="store_false", help="Don't save history in cache")
    parser.set_defaults(use_cache=True)
    parser.add_argument(
        "--use-notify-send",
        dest="use_notify_send",
        action="store_true",
        help="Send desktop notification after copying TOTP",
    )
    parser.set_defaults(use_notify_send=False)
    parser.add_argument(
        "--keybindings",
        dest="keybindings",
        action="store",
        type=str,
        default=",".join(
            [
                "Alt+1:type:username:tab:password",
                "Alt+2:type:username",
                "Alt+3:type:password",
                "Alt+4:type:totp",
                "Alt+c:copy:password",
                "Alt+u:copy:username",
                "Alt+t:copy:totp",
                "Alt+m::menu",
                "Alt+s:sync",
            ]
        ),
        help="Define keyboard shortcuts in the format <shortcut>:<action>:<target>, separated with a comma",
    )
    parser.add_argument(
        "--menu-keybindings",
        dest="menu_keybindings",
        action="store",
        type=str,
        default=",".join(["Alt+t:type", "Alt+c:copy"]),
        help="Define the keyboard shortcuts in the target menu in the format <shortcut>:<action> separated with a comma",
    )
    parser.add_argument(
        "--typing-key-delay",
        dest="key_delay",
        action="store",
        type=int,
        default=0,
        help="Set the delay between key presses when typing",
    )
    parser.add_argument(
        "--typing-start-delay",
        dest="start_delay",
        action="store",
        type=int,
        default=0,
        help="Set the typing start delay",
    )
    parser.add_argument(
        "--action-sequence-delay",
        dest="action_sequence_delay",
        action="store",
        type=int,
        default=1000,
        help="Set the delay to wait for a 'delay' step",
    )

    parsed_args = parser.parse_args(argv)

    parsed_args.selector_args = shlex.split(parsed_args.selector_args)

    parsed_args.action = Action(parsed_args.action)

    if parsed_args.targets:
        parsed_args.targets = [Target(target) for target in parsed_args.targets]
    else:
        parsed_args.targets = None

    parsed_args.parsed_keybindings = []
    for keybinding in parsed_args.keybindings.split(","):
        elements = keybinding.split(":")
        parsed_args.parsed_keybindings.append(
            Keybinding(
                elements[0],
                Action(elements[1]) if elements[1] else None,
                [TypeTarget(target_string) for target_string in elements[2:]],
            )
        )

    parsed_args.parsed_menu_keybindings = []
    for keybinding in parsed_args.menu_keybindings.split(","):
        elements = keybinding.split(":")
        parsed_args.parsed_menu_keybindings.append(
            Keybinding(elements[0], Action(elements[1]) if elements[1] else None, None)
        )

    if parsed_args.display_fields is not None:
        display_fields = [DisplayFieldToken(field) for field in parsed_args.display_fields]
    elif not parsed_args.show_folders:
        sys.stderr.write(
            "warning: --no-folder is deprecated and will be removed in a future release. Use --display-fields name_only user instead.\n"
        )
        display_fields = [DisplayFieldToken.NAME_ONLY, DisplayFieldToken.USER]
    else:
        display_fields = [DisplayFieldToken.NAME_WITH_FOLDER, DisplayFieldToken.USER]

    del parsed_args.show_folders
    parsed_args.display_fields = display_fields

    return parsed_args



import json
from json import JSONDecodeError
from subprocess import run
from typing import Any

from .models.card import Card
from .models.credentials import Credentials
from .models.detailed_entry import DetailedEntry
from .models.entry import Entry
from .models.EntryType import EntryType
from .models.field import Field, FieldType
from .models.note import Note


class Rbw:
    def list_entries(self) -> list[Entry]:
        rbw = run(["rbw", "list", "--raw"], encoding="utf-8", capture_output=True)

        if rbw.returncode != 0:
            print("There was a problem calling rbw. Is it correctly configured?")
            print(rbw.stderr)
            exit(2)

        data = json.loads(rbw.stdout.strip())

        return sorted(
            [
                Entry(item["name"], item["folder"] or "", item["user"] or "", item["type"], item["uris"] or [])
                for item in data
                if item["type"] in {EntryType.LOGIN.value, EntryType.CARD.value, EntryType.NOTE.value}
            ],
            key=lambda x: x.folder.lower() + x.name.lower(),
        )

    def fetch_credentials(self, entry: Entry) -> DetailedEntry:
        match entry.type:
            case EntryType.LOGIN.value:
                return self.__fetch_login(entry)
            case EntryType.CARD.value:
                return self.__fetch_card(entry)
            case EntryType.NOTE.value:
                return self.__fetch_note(entry)
            case _:
                print(f"Unsupported type: {entry.type}")
                exit(7)

    def __fetch_login(self, entry: Entry) -> Credentials:
        data = self.__load_from_rbw(entry.name, entry.username, entry.folder)

        return Credentials(
            entry.name,
            data["folder"],
            [Field(item["name"], item["value"], FieldType(item["type"])) for item in data["fields"] if item["name"]],
            entry.username,
            data["data"]["password"] or "",
            data["data"]["totp"] is not None,
            data["notes"],
            [item["uri"] for item in data["data"]["uris"]],
        )

    def __fetch_card(self, entry: Entry) -> Card:
        data = self.__load_from_rbw(entry.name, entry.username, entry.folder)

        return Card(
            entry.name,
            entry.folder,
            [Field(item["name"], item["value"], FieldType(item["type"])) for item in data["fields"] if item["name"]],
            data["data"]["cardholder_name"],
            data["data"]["number"],
            data["data"]["brand"],
            data["data"]["exp_month"],
            data["data"]["exp_year"],
            data["data"]["code"],
            data["notes"],
        )

    def __fetch_note(self, entry: Entry) -> Note:
        data = self.__load_from_rbw(entry.name, entry.username, entry.folder)

        return Note(
            entry.name,
            entry.folder,
            [Field(item["name"], item["value"], FieldType(item["type"])) for item in data["fields"] if item["name"]],
            data["notes"],
        )

    def __load_from_rbw(self, name: str, username: str, folder: str | None) -> dict[str, Any]:
        command = ["rbw", "get", "--raw", name]
        if username:
            command.append(username)

        if folder:
            command.extend(["--folder", folder])

        data = run(command, capture_output=True, encoding="utf-8").stdout

        try:
            return json.loads(data.strip())

        except JSONDecodeError as exception:
            print(f"Could not parse the output: {exception.msg}")
            exit(12)

    def sync(self):
        run(["rbw", "sync"])



from .typer import Key, NoTyperFoundException, Typer


class NoopTyper(Typer):
    @staticmethod
    def supported() -> bool:
        return True

    @staticmethod
    def name() -> str:
        return "noop"

    def get_active_window(self) -> str:
        raise NoTyperFoundException()

    def type_characters(self, characters: str, start_delay: float, key_delay: int, active_window: str) -> None:
        raise NoTyperFoundException()

    def press_key(self, key: Key) -> None:
        raise NoTyperFoundException()



from dataclasses import dataclass, field
from functools import cached_property
from hashlib import sha1

from rofi_rbw.models.EntryType import EntryType


@dataclass(frozen=True)
class Entry:
    name: str
    folder: str
    username: str
    type: EntryType
    uris: list[str] = field(default_factory=list)

    @cached_property
    def hashed(self) -> str:
        m = sha1()
        m.update(self.name.encode())
        if self.folder:
            m.update(self.folder.encode())
        if self.username:
            m.update(self.username.encode())
        return m.hexdigest()



from subprocess import run

from ..abstractionhelper import is_installed
from ..models.action import Action
from ..models.detailed_entry import DetailedEntry
from ..models.display_field_token import DisplayFieldToken
from ..models.entry import Entry
from ..models.keybinding import Keybinding
from ..models.targets import Target
from .selector import Selector


class Rofi(Selector):
    @staticmethod
    def supported() -> bool:
        return is_installed("rofi")

    @staticmethod
    def name() -> str:
        return "rofi"

    def show_selection(
        self,
        entries: list[Entry],
        prompt: str,
        show_help_message: bool,
        display_fields: list[DisplayFieldToken],
        keybindings: list[Keybinding],
        additional_args: list[str],
    ) -> tuple[list[Target] | None, Action | None, Entry | None]:
        parameters = [
            "rofi",
            "-markup-rows",
            "-dmenu",
            "-format",
            "i",
            "-i",
            "-sort",
            "-p",
            prompt,
            *self.__build_parameters_for_keybindings(keybindings),
            *additional_args,
        ]

        if show_help_message and keybindings:
            parameters.extend(self.__format_keybindings_message(keybindings))

        rofi = run(
            parameters,
            input="\n".join(self._format_entries(entries, display_fields)),
            capture_output=True,
            encoding="utf-8",
        )

        if rofi.returncode == 1:
            return None, Action.CANCEL, None
        elif rofi.returncode >= 10:
            keybinding = keybindings[(rofi.returncode - 10)]
            return_action = keybinding.action
            return_targets = keybinding.targets
        else:
            return_action = None
            return_targets = None

        return return_targets, return_action, (entries[int(rofi.stdout.strip())])

    def _format_field(self, entry: Entry, token: DisplayFieldToken) -> str:
        match token:
            case DisplayFieldToken.NAME_ONLY:
                return f"<b>{super()._format_field(entry, token)}</b>"
            case DisplayFieldToken.NAME_WITH_FOLDER:
                if entry.folder:
                    return f"{entry.folder}/<b>{entry.name}</b>"
                return f"<b>{entry.name}</b>"
            case _:
                return super()._format_field(entry, token)

    def select_target(
        self,
        entry: DetailedEntry,
        show_help_message: bool,
        keybindings: list[Keybinding],
        additional_args: list[str],
    ) -> tuple[list[Target] | None, Action | None]:
        parameters = [
            "rofi",
            "-markup-rows",
            "-dmenu",
            "-p",
            "Choose target",
            "-i",
            *self.__build_parameters_for_keybindings(keybindings),
            *additional_args,
        ]

        if show_help_message and keybindings:
            parameters.extend(self.__format_keybindings_message(keybindings))

        rofi = run(
            parameters,
            input="\n".join(self._format_targets_from_entry(entry)),
            capture_output=True,
            encoding="utf-8",
        )

        if rofi.returncode == 1:
            return None, Action.CANCEL
        elif rofi.returncode >= 10:
            action = keybindings[(rofi.returncode - 10)].action
        else:
            action = None

        return (self._extract_targets(rofi.stdout)), action

    def __build_parameters_for_keybindings(self, keybindings: list[Keybinding]) -> list[str]:
        params = []
        for index, keybinding in enumerate(keybindings):
            params.extend([f"-kb-custom-{1 + index}", keybinding.shortcut])

        return params

    def __format_keybindings_message(self, keybindings: list[Keybinding]):
        return [
            "-mesg",
            " | ".join(
                [
                    f"<b>{keybinding.shortcut}</b>: {self._format_action_and_targets(keybinding)}"
                    for keybinding in keybindings
                ]
            ),
        ]



import os
import shutil


def is_installed(executable: str) -> bool:
    return shutil.which(executable) is not None


def is_wayland() -> bool:
    return os.environ.get("WAYLAND_DISPLAY") is not None



import time
from subprocess import run

from ..abstractionhelper import is_installed, is_wayland
from .clipboarder import Clipboarder


class WlClipboarder(Clipboarder):
    __last_copied_characters: str | None

    @staticmethod
    def supported() -> bool:
        return is_wayland() and is_installed("wl-copy")

    @staticmethod
    def name() -> str:
        return "wl-copy"

    def copy_to_clipboard(self, characters: str) -> None:
        run(["wl-copy", "--sensitive"], input=characters, encoding="utf-8")

        self.__last_copied_characters = characters

    def clear_clipboard_after(self, clear: int) -> None:
        if clear > 0:
            time.sleep(clear)

            if self.__fetch_clipboard_content() == self.__last_copied_characters:
                run(["wl-copy", "--clear"])
                self.__last_copied_characters = None

    def __fetch_clipboard_content(self) -> str:
        return run(["wl-paste", "-n"], capture_output=True, encoding="utf-8").stdout



from dataclasses import dataclass
from enum import Enum


class FieldType(Enum):
    HIDDEN = "hidden"
    TEXT = "text"
    BOOLEAN = "boolean"
    LINKED = "linked"


@dataclass
class Field:
    key: str
    value: str
    type: FieldType | None

    def masked_string(self):
        if self.type == FieldType.HIDDEN:
            return self.value[0] + ("*" * (len(self.value) - 1))

        return self.value



from subprocess import run


class Notifier:
    def notify(self, message: str) -> None:
        run(
            [
                "notify-send",
                "-u",
                "normal",
                "rofi-rbw",
                message,
            ],
        )



from math import floor

from .models.entry import Entry
from .paths import cache_file


class Cache:
    def __init__(self):
        cache_file.parent.mkdir(exist_ok=True)
        self.cache = dict()

    def sorted(self, entries: list[Entry]) -> list[Entry]:
        hashed_entries = {entry.hashed: entry for entry in entries}
        sorted_entries = []

        if cache_file.exists():
            with cache_file.open() as f:
                for line in f:
                    stripped_line = line.strip()
                    if stripped_line:
                        amount, hashed = stripped_line.split(" ", maxsplit=1)
                        try:
                            sorted_entries.append(hashed_entries.pop(hashed))
                            self.cache[hashed] = int(amount)
                        except KeyError:
                            pass

        return [*sorted_entries, *hashed_entries.values()]

    def update(self, entry: Entry):
        self.cache[entry.hashed] = self.cache.get(entry.hashed, 0) + 1.1

        with cache_file.open("w") as f:
            for hashed, n in sorted(self.cache.items(), key=lambda i: i[1], reverse=True):
                f.write(f"{floor(n)} {hashed}\n")



import time

from .argument_parsing import parse_arguments
from .cache import Cache
from .clipboarder.clipboarder import Clipboarder
from .models.action import Action
from .models.credentials import Credentials
from .models.detailed_entry import DetailedEntry
from .models.targets import Target, Targets, TypeTargets
from .notifier import Notifier
from .rbw import Rbw
from .selector.selector import Selector
from .typer.typer import Key, Typer


class RofiRbw(object):
    def __init__(self) -> None:
        self.args = parse_arguments()
        self.rbw = Rbw()
        self.selector = Selector.best_option(self.args.selector)
        self.typer = Typer.best_option(self.args.typer)
        self.clipboarder = Clipboarder.best_option(self.args.clipboarder)
        self.notifer = Notifier()
        self.active_window = self.typer.get_active_window()

    def main(self) -> None:
        entries = self.rbw.list_entries()

        if self.args.use_cache:
            cache = Cache()
            entries = cache.sorted(entries)

        (selected_targets, selected_action, selected_entry) = self.selector.show_selection(
            entries,
            self.args.prompt,
            self.args.show_help,
            self.args.display_fields,
            self.args.parsed_keybindings,
            self.args.selector_args,
        )

        if selected_action == Action.SYNC:
            self.rbw.sync()
            (selected_targets, selected_action, selected_entry) = self.selector.show_selection(
                self.rbw.list_entries(),
                self.args.prompt,
                self.args.show_help,
                self.args.display_fields,
                self.args.parsed_keybindings,
                self.args.selector_args,
            )

        if selected_action == Action.CANCEL:
            return

        entry = self.rbw.fetch_credentials(selected_entry)

        if self.args.use_cache:
            cache.update(selected_entry)

        if selected_targets is not None:
            self.args.targets = selected_targets

        if selected_action is not None:
            self.args.action = selected_action

        if self.args.targets is not None and Targets.MENU in self.args.targets:
            targets, action = self.__show_target_menu(
                entry,
                self.args.show_help,
            )
            self.args.targets = targets
            if action is not None:
                self.args.action = action

        self.__execute_action(entry)

    def __show_target_menu(self, entry: DetailedEntry, show_help_message: bool) -> tuple[list[Target], Action | None]:
        targets, action = self.selector.select_target(
            entry, show_help_message, self.args.parsed_menu_keybindings, additional_args=self.args.selector_args
        )

        if action == Action.CANCEL:
            self.main()
            exit(0)

        return targets, action

    def __execute_action(self, detailed_entry: DetailedEntry) -> None:
        targets = self.__configure_targets(detailed_entry)
        match self.args.action:
            case Action.TYPE:
                self.__type_targets(detailed_entry, targets)
            case Action.COPY:
                self.__copy_targets(detailed_entry, targets)
            case Action.PRINT:
                print("\n".join([detailed_entry[target] for target in targets]))

    def __configure_targets(self, detailed_entry: DetailedEntry) -> list[Target]:
        if self.args.targets:
            return self.args.targets

        if self.args.action == Action.TYPE:
            return detailed_entry.autotype_sequence or detailed_entry.default_autotype_target

        return detailed_entry.default_target

    def __type_targets(self, detailed_entry: DetailedEntry, targets: list[Target]):
        for target in targets:
            match target:
                case TypeTargets.DELAY:
                    time.sleep(self.args.action_sequence_delay / 1000)
                case TypeTargets.ENTER:
                    self.typer.press_key(Key.ENTER)
                case TypeTargets.TAB:
                    self.typer.press_key(Key.TAB)
                case _:
                    self.typer.type_characters(
                        detailed_entry[target], self.args.start_delay, self.args.key_delay, self.active_window
                    )
        if Targets.PASSWORD in targets and isinstance(detailed_entry, Credentials) and detailed_entry.totp != "":
            self.clipboarder.copy_to_clipboard(detailed_entry.totp)
            if self.args.use_notify_send:
                self.notifer.notify("totp copied to clipboard")

    def __copy_targets(self, detailed_entry: DetailedEntry, targets: list[Target]):
        for target in targets:
            if target == TypeTargets.DELAY:
                time.sleep(self.args.action_sequence_delay / 1000)
            else:
                value = detailed_entry[target]
                self.clipboarder.copy_to_clipboard(value)

                if self.args.use_notify_send:
                    self.notifer.notify(f"{target.raw} copied to clipboard")

        if len(targets) == 1 and targets[0] == Targets.PASSWORD:
            self.clipboarder.clear_clipboard_after(self.args.clear)



from dataclasses import dataclass

from .action import Action
from .targets import TypeTarget


@dataclass
class Keybinding:
    shortcut: str
    action: Action | None
    targets: list[TypeTarget] | None
